/* {{{ Copyright etc. */

/*  gfeat - various group stats analyses

    Stephen Smith and Didier Leibovici, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    
    LICENCE
    
    FMRIB Software Library, Release 3.0 (c) 2002, The University of Oxford
    (the "Software")
    
    The Software remains the property of the University of Oxford ("the
    University").
    
    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.
    
    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.
    
    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.
    
    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Isis
    Innovation Limited ("Isis"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    innovation@isis.ox.ac.uk quoting reference DE/1112. */

/* }}} */
/* {{{ equations used */

/*

The general form of output produced before calling ttoz is a numerator
and denominator for the t values: i.e. t = PE / sqrt(V)

PE  = some kind of group parameter estimate
V   = some kind of group variance estimate
DOF = the relevant degrees of freedom for each group analysis

So - the equations for these 3 quantities are:

PE_i    = parameter estimate (or contrast of PEs) from subject i
V_i     = estimate of variance of PE_i
DOF_i   = the DOF for subject i, i.e. the corrected number of time points
N_a     = number of subjects in group a
          (control group = is zero for single-group tests)
SUM_a() = sum over group a
N_b     = number of subjects in group b (test group)
SUM_b() = sum over group b

For all methods the PE is the same, whilst V and DOF vary:

     SUM_b(PE_i)   SUM_a(PE_i)
PE = ----------- - -----------
         N_b           N_a

**** FIXED EFFECTS **********************************

    SUM_a(V_i)   SUM_b(V_i)
V = ---------- + ----------
      N_a^2        N_b^2

DOF = SUM_a&b(DOF_i)

**** RANDOM EFFECTS *********************************

Unbiased estimators:

                         N_a    Sum_a( (PE_i-Mean_a(PE_i))^2 ) 
V_a = Variance(PE_ia) = ----- * ------------------------------
                        N_a-1               N_a

      SUM_a(PE_i^2)   SUM_a(PE_i)^2
    = ------------- - -------------
          N_a-1        N_a*(N_a-1)

(same for V_b)

    V_a   V_b
V = --- + ---
    N_a   N_b

    SUM_a(PE_i^2)   SUM_a(PE_i)^2   SUM_b(PE_i^2)   SUM_b(PE_i)^2
  = ------------- - ------------- + ------------- - -------------
     N_a*(N_a-1)    N_a^2*(N_a-1)    N_b*(N_b-1)    N_b^2*(N_b-1)

DOF = N_a + N_b - 2

**** Worsley's RATIO ********************************

V = Vfe * blurred( Vre / Vfe )

    where blurred means spatially blurred image of the variance
    ratio by a prespecified amount

DOF = 1 / ( 1/DOFratio + 1/DOFfe )
    
      where DOFratio = DOFrandom * (2*(FWHMblur/FWHMresel)^2+1)^1.5

*/

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss.h"
#include "libavw.h"

void usage(void);

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("Usage: gfeat4 <Na> <Nb> <a1.feat> <a2.feat> ..... <b1.feat> <b2.feat> ..... <outdir> [-r fwhm_blur fwhm_resel> [-p pluses] [-k] [-n]\n");
  printf("Set Na to 0 for one-group tests. All two-group tests are based on B-A.\n");
  printf("The [pluses] string is for selecting re-run contrast outputs, e.g. ++\n");
  printf("-k : keep intermediate mean diff and variance images\n");
  printf("-n : no registration (already in standard space)\n");
  exit(1);
}

/* }}} */
/* {{{ main */

int main(int argc, char **argv)
{
  /* {{{ variables */

FILE          *fp;
FDT           *in, *PE=NULL, *Vfe=NULL, *Vre=NULL, *tmp=NULL;
image_struct  im;
int           i, j, size=0, argi=1, Na, Nb, con, mean_highres=1, keep=0, doreg=1;
char          astring[10000], pluses[10000]={0}, outdir[10000], binstr[10000], standard[10000], highres[10000];
float         tmpf=0, sumdof=0, doffe, dofre, dofratio, fwhm_blur=-1, fwhm_resel=0;
unsigned char *mask=NULL;

/* }}} */

  /* {{{ basic usage and Na and Nb */

if (argc<3)
     usage();

Na=atoi(argv[argi++]);
Nb=atoi(argv[argi++]);

if ( (Na<0) || (Na==1) || (Nb<2) || (argc<4+Na+Nb) )
     usage();

/* will use binstr to work out where "gnuplot" and "convert" are, for later system calls */
sprintf(binstr,"%s",argv[0]);
binstr[strlen(binstr)-strlen("gfeat4")]=0;

/* }}} */
  /* {{{ find sumdof */

for(i=0; i<Na+Nb; i++)
{
  sprintf(astring,"%s/stats/dof",argv[argi]);

  if((fp=fopen(astring,"r"))==NULL)
    {
      printf("Can't open %s\n",astring);
      exit(1);
    }

  fscanf(fp,"%f",&tmpf);
  printf("dof[%d]=%f\n",i,tmpf);
  sumdof+=tmpf;

  fclose(fp);

  argi++;
}

/* }}} */
  /* {{{ setup output directory */

strcpy(outdir,argv[argi++]);

/* }}} */
  /* {{{ process optional inputs */

for (; argi < argc; argi++)
{
  if (!strcmp(argv[argi], "-r"))
    {
      fwhm_blur  = atof(argv[++argi]);
      fwhm_resel = atof(argv[++argi]);
      printf("fwhm_blur=%f fwhm_resel=%f\n",fwhm_blur,fwhm_resel);
    }
  else if (!strcmp(argv[argi], "-p"))
    {
      strcpy(pluses,argv[++argi]);
      printf("pluses[%s]\n",pluses);
    }
  else if (!strcmp(argv[argi], "-k"))
    keep=1;
  else if (!strcmp(argv[argi], "-n"))
    doreg=0;
}

/* }}} */
  /* {{{ find standard image */

if (doreg)
{
  sprintf(standard,"%s/standard.hdr",argv[3]); /* first FEAT directory's standard image - assume all the same */
  if ((fp=fopen(standard,"r"))==NULL)
    {
      printf("Can't open standard image %s for reading.\n",standard);
      exit(1);
    }
  fclose(fp);
  
  standard[strlen(standard)-4]=0;
  
  printf("standard image = %s\n",standard);
}

/* }}} */

  sprintf(astring,"%s/stats/cope%s1.hdr",argv[3],pluses);
  if((fp=fopen(astring,"r"))!=NULL)
    /* {{{ use FILM (cope/varcope) output */

{
  fclose(fp);
  for(con=0; ;con++)
    {
      for(i=0, argi=3; i<Na+Nb; i++, argi++)
	{
	  /* {{{ test for existence of cope image */

sprintf(astring,"%s/stats/cope%s%d.hdr",argv[argi],pluses,con+1);

if((fp=fopen(astring,"r"))==NULL)
{
  if ((con==0)||(i>0))
    {
      printf("Whooops - can't find %s!\n",astring);
      exit(1);
    }
  else
    {
      /* {{{ make mean highres */

if (mean_highres && doreg)
{
  int hi, hargi;

  memset(tmp,  0, sizeof(float)*size);

  for(hi=0, hargi=3; hi<Na+Nb; hi++, hargi++)
    /* {{{ transform and load highres image */

{
  sprintf(highres,"%s/highres.hdr",argv[hargi]);
  if ((fp=fopen(highres,"r"))==NULL)
    {
      printf("Can't open high res image %s for reading\n",highres);
      hi=Na+Nb+1; /* break out of loops and dont make mean highres */
      mean_highres=0;
    }
  else
    {
      image_struct imh;

      fclose(fp);
      highres[strlen(highres)-4]=0;
      printf("highres image = %s\n",highres);

      sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/highres2standard.mat -in %s -ref %s",
	      binstr,outdir,argv[hargi],highres,standard);
      printf("%s\n",astring);
      system(astring);
      sprintf(astring,"%s/tmp",outdir);
      avw_read(astring,&imh);
      in=imh.i;

      for(j=0; j<size; j++)
	tmp[j] += in[j];

      free(in);
    }
}

/* }}} */
  
  if (mean_highres)
    {
      for(j=0; j<size; j++)
	tmp[j] /= (Na+Nb);
      
      sprintf(astring,"%s/mean_highres",outdir);
      im.i=tmp;
      avw_write(astring,im);
    }
}

/* }}} */
      /* {{{ clean up temporary images */

if (!keep)
{
  sprintf(astring,"rm -rf %s/V* %s/PE* %s/tmp*",outdir,outdir,outdir);
  printf("%s\n",astring);
  system(astring);
}

/* }}} */
      return 0; /* terminate program as we've done all of the contrasts */
    }
}
fclose(fp);

/* }}} */
	  /* {{{ transform and load cope image */

if (doreg)
{
  sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/example_func2standard.mat -in %s/stats/cope%s%d -ref %s",
	  binstr,outdir,argv[argi],argv[argi],pluses,con+1,standard);
  printf("%s\n",astring);
  system(astring);
  sprintf(astring,"%s/tmp",outdir);
  avw_read(astring,&im);
} else {
  sprintf(astring,"%s/stats/cope%s%d",argv[argi],pluses,con+1);
  avw_read(astring,&im);
}

in=im.i;

/* }}} */
	  /* {{{ if first load, malloc working images */

if ((i==0)&&(con==0))
{
  size = im.x*im.y*im.z;
  PE   = malloc(sizeof(float)*size);
  Vfe  = malloc(sizeof(float)*size);
  Vre  = malloc(sizeof(float)*size);
  tmp  = malloc(sizeof(float)*size);
  mask = malloc(sizeof(unsigned char)*size);
}
else if (size!=im.x*im.y*im.z)
{
  printf("Grots - not all the transformed input images are the same size!\n");
  exit(1);
}

if (i==0)
{
  memset(PE,   0, sizeof(float)*size);
  memset(Vfe,  0, sizeof(float)*size);
  memset(Vre,  0, sizeof(float)*size);
  memset(tmp,  0, sizeof(float)*size);
  memset(mask, 1, sizeof(unsigned char)*size);
}

/* }}} */
	  /* {{{ add to cope sums */

for(j=0; j<size; j++)
{
  tmpf = in[j];

  if (i<Na)
    {
      PE[j]  -= tmpf/Na;
      Vre[j] += tmpf*tmpf/(Na*(Na-1));
      tmp[j] += tmpf;
    }
  else
    {
      PE[j]  += tmpf/Nb;
      Vre[j] += tmpf*tmpf/(Nb*(Nb-1));
      tmp[j] += tmpf;
    }
}

free(in);

if ( (Na>0) && (i==Na-1) )
{
  for(j=0; j<size; j++)
    Vre[j] -= tmp[j]*tmp[j]/(Na*Na*(Na-1));
  memset(tmp, 0,sizeof(float)*size);
}
if (i==Na+Nb-1)
  for(j=0; j<size; j++)
    Vre[j] -= tmp[j]*tmp[j]/(Nb*Nb*(Nb-1));

/* }}} */
	  /* {{{ transform and load variance image */

if (doreg)
{
  sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/example_func2standard.mat -in %s/stats/varcope%s%d -ref %s",
	  binstr,outdir,argv[argi],argv[argi],pluses,con+1,standard);
  printf("%s\n",astring);
  system(astring);
  sprintf(astring,"%s/tmp",outdir);
  avw_read(astring,&im);
} else {
  sprintf(astring,"%s/stats/varcope%s%d",argv[argi],pluses,con+1);
  avw_read(astring,&im);
}

in=im.i;

/* }}} */
	  /* {{{ add to Vfe sum */

for(j=0; j<size; j++)
{
  if (i<Na)
    Vfe[j] += in[j]/(Na*Na);
  else
    Vfe[j] += in[j]/(Nb*Nb);

  if (in[j]==0)
    mask[j]=0;
}

free(in);

/* }}} */
	}

      /* {{{ apply and write mask */

for(j=0; j<size; j++)
  if (mask[j]==0)
    PE[j]=0;

if (con==0)
{
  int tmpdt=im.dt, tmpbpv=im.bpv;
  sprintf(astring,"%s/mask",outdir);
  im.i=(FDT*)mask;
  im.dt=DT_UNSIGNED_CHAR;
  im.bpv=sizeof(unsigned char);
  im.thresh2=im.min=0;
  im.thresh98=im.max=1;
  avw_write(astring,im);
  im.dt=tmpdt;
  im.bpv=tmpbpv;
}

/* }}} */
      /* {{{ output FE images and compute Z */

sprintf(astring,"%s/PE%d",outdir,con+1);
im.i=PE;
avw_write(astring,im);

/* create example AVW header file */
sprintf(astring,"cp %s/PE1.hdr %s/.hdr",outdir,outdir);
printf("%s\n",astring);
system(astring);

sprintf(astring,"%s/fV%d",outdir,con+1);
im.i=Vfe;
avw_write(astring,im);

doffe=sumdof;
printf("fixed-effects group dof = %f\n",doffe);

sprintf(astring,"%sttoz -zout %s/fixed%d %s/fV%d %s/PE%d %f",binstr,outdir,con+1,outdir,con+1,outdir,con+1,doffe);
printf("%s\n",astring);
system(astring);

/* }}} */
      /* {{{ output RE images and compute Z */

sprintf(astring,"%s/rV%d",outdir,con+1);
im.i=Vre;
avw_write(astring,im);

if (Na==0)
     dofre=Nb-1;
else
     dofre=Na+Nb-2;
printf("random-effects group dof = %f\n",dofre);

sprintf(astring,"%sttoz -zout %s/random%d %s/rV%d %s/PE%d %f",binstr,outdir,con+1,outdir,con+1,outdir,con+1,dofre);
printf("%s\n",astring);
system(astring);

/* }}} */
      /* {{{ output Worsley images and compute Z */

if (fwhm_blur>0)
{
  for(j=0; j<size; j++)
    {
      if (Vfe[j]>0)
	Vre[j] /= Vfe[j];
      else
	Vre[j]=0;
    }
  
  im.i=Vre;
  spatfilt(im,fwhm_blur/2.355); /* convert FWHM to sigma */
  
  for(j=0; j<size; j++)
    Vre[j] *= Vfe[j];     

  sprintf(astring,"%s/wV%d",outdir,con+1);
  avw_write(astring,im);
  
  dofratio= 1 / ( 1/(dofre*pow(2*(fwhm_blur/fwhm_resel)*(fwhm_blur/fwhm_resel)+1,1.5)) + 1/doffe );
  printf("Ratio dof = %f\n",dofratio);
  
  sprintf(astring,"%sttoz -zout %s/ratio%d %s/wV%d %s/PE%d %f",binstr,outdir,con+1,outdir,con+1,outdir,con+1,dofratio);
  printf("%s\n",astring);
  system(astring);
}

/* }}} */
    }
}

/* }}} */
  else
    /* {{{ use IRVA (zfstat) output */

{
  int donehighres=0;

  for(i=0, argi=3; i<Na+Nb; i++, argi++)
    {
      /* {{{ transform and load zfstat image */

if (doreg)
{
  sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/example_func2standard.mat -in %s/stats/zfstat%s%d -ref %s",
	  binstr,outdir,argv[argi],argv[argi],pluses,1,standard);
  printf("%s\n",astring);
  system(astring);
  sprintf(astring,"%s/tmp",outdir);
  avw_read(astring,&im);
} else {
  sprintf(astring,"%s/stats/zfstat%s%d",argv[argi],pluses,1);
  avw_read(astring,&im);
}

in=im.i;

/* }}} */
      /* {{{ if first load, malloc working images */

if (i==0)
{
  size = im.x*im.y*im.z;
  PE   = malloc(sizeof(float)*size);
  tmp  = malloc(sizeof(float)*size);
  mask = malloc(sizeof(unsigned char)*size);
  memset(PE,   0, sizeof(float)*size);
  memset(tmp,  0, sizeof(float)*size);
  memset(mask, 1, sizeof(unsigned char)*size);
}

/* }}} */
      /* {{{ add to zfstat sum */

for(j=0; j<size; j++)
{
  tmpf = in[j];

  if (i<Na)
    PE[j]  -= tmpf/Na;
  else
    PE[j]  += tmpf/Nb;
}

free(in);

/* }}} */
      /* {{{ transform, load and use mask image */

if (doreg)
{
  sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/example_func2standard.mat -in %s/mask -ref %s",
	  binstr,outdir,argv[argi],argv[argi],standard);
  printf("%s\n",astring);
  system(astring);
  sprintf(astring,"%s/tmp",outdir);
  avw_read(astring,&im);
} else {
  sprintf(astring,"%s/mask",argv[argi]);
  avw_read(astring,&im);
}

in=im.i;

for(j=0; j<size; j++)
     mask[j] *= in[j];

free(in);

/* }}} */
      /* {{{ transform, load and use standard image */

if (doreg)
{
sprintf(astring,"%s/highres.hdr",argv[argi]);

if((fp=fopen(astring,"r"))!=NULL)
{
  fclose(fp);
  sprintf(astring,"%sflirt -out %s/tmp -applyxfm -init %s/highres2standard.mat -in %s/highres -ref %s",
	  binstr,outdir,argv[argi],argv[argi],standard);
  printf("%s\n",astring);
  system(astring);
  sprintf(astring,"%s/tmp",outdir);
  avw_read(astring,&im);
  in=im.i;

  for(j=0; j<size; j++)
    tmp[j] += in[j];

  free(in);

  donehighres=1;
}
}

/* }}} */
    }

  /* {{{ apply and write mask */

for(j=0; j<size; j++)
  if (mask[j]==0)
    PE[j]=0;

{
  int tmpdt=im.dt, tmpbpv=im.bpv;
  sprintf(astring,"%s/mask",outdir);
  im.i=(FDT*)mask;
  im.dt=DT_UNSIGNED_CHAR;
  im.bpv=sizeof(unsigned char);
  im.thresh2=im.min=0;
  im.thresh98=im.max=1;
  avw_write(astring,im);
  im.dt=tmpdt;
  im.bpv=tmpbpv;
}

/* }}} */
  /* {{{ write standard */

if (donehighres)
{
  int tmpdt=im.dt, tmpbpv=im.bpv;
  short *st=malloc(size*sizeof(short));

  for(j=0; j<size; j++)
    st[j]=tmp[j]/(Na+Nb);

  sprintf(astring,"%s/mean_highres",outdir);
  im.i=(FDT*)st;
  im.dt=DT_SIGNED_SHORT;
  im.bpv=sizeof(short);
  avw_write(astring,im);
  im.dt=tmpdt;
  im.bpv=tmpbpv;
}

/* }}} */
  /* {{{ renorm and output Z stat image */

for(j=0; j<size; j++)
     PE[j] *= sqrt(Na+Nb); /* not /= sqrt() because we've already divided Na/b once */

sprintf(astring,"%s/zstat1",outdir);
im.i=PE;
avw_write(astring,im);

/* }}} */
  /* {{{ clean up temporary images */

  sprintf(astring,"rm -rf %s/tmp*",outdir);
  printf("%s\n",astring);
  system(astring);

/* }}} */

  return 0;
}

/* }}} */
}

/* }}} */
