#if !defined(__complexvolume_h)
#define __complexvolume_h

#include "newimageall.h"

namespace NEWIMAGE {

  class complexpoint;

  class complexref {
  private:
    float *m_real;
    float *m_imag;
  public:
    complexref(float* r, float* i) : m_real(r), m_imag(i){}
    ~complexref(){}
    inline float* re_pointer() const { return m_real; }
    inline float* im_pointer() const { return m_imag; }
    const complexpoint& operator=(const complexpoint& val);
  };

  class complexpoint {

  private:
    float m_real;
    float m_imag;

  public:

    complexpoint(){}
    complexpoint(float r, float i){m_real=r; m_imag=i;}
    complexpoint(float r){m_real=r; m_imag=0;}
    complexpoint(const complexref& source){
      m_real = *(source.re_pointer());
      m_imag = *(source.im_pointer());
    }
    ~complexpoint(){}
    float operator=(const float val);
    complexpoint& operator=(const complexpoint& source);
    complexpoint& operator=(const complexref& source);
    inline float re() const { return m_real; }
    inline float im() const { return m_imag; }

    const complexpoint& operator+=(const complexpoint& val); 
    const complexpoint& operator-=(const complexpoint& val); 
    const complexpoint& operator*=(const complexpoint& val); 
    const complexpoint& operator/=(const complexpoint& val);
    complexpoint operator+(const complexpoint& val) const;
    complexpoint operator-(const complexpoint& val) const;
    complexpoint operator*(const complexpoint& val) const;
    complexpoint operator/(const complexpoint& val) const;

    //ostream& operator<<(ostream& s, const complexpoint& val);

    float abs() const;
    float phase() const;
  };

  class complexvolume {
    
  private:
    volume<float> real;
    volume<float> imag;

  public:
    complexvolume(){}
    complexvolume(const complexvolume& source);
    complexvolume(const volume<float>& r, const volume<float>& i);
    complexvolume(const volume<float>& r);
    ~complexvolume();
    float operator=(const float val);
    complexvolume& operator=(const complexvolume& source); 
    void destroy();
    int copyproperties(const complexvolume& source);
    int copydata(const complexvolume& source);
 
    float re(int x,int y, int z) const { return real(x,y,z); }
    float im(int x,int y, int z) const { return imag(x,y,z); }

    inline int xsize() const { return real.xsize(); }
    inline int ysize() const { return real.ysize(); }
    inline int zsize() const { return real.zsize(); }
    inline float xdim() const { return real.xdim(); }
    inline float ydim() const { return real.ydim(); }
    inline float zdim() const { return real.zdim(); }
    inline const int xsign(void) const { return real.xsign(); }
    inline const int ysign(void) const { return real.zsign(); }
    inline const int zsign(void) const { return real.ysign(); }
    void setxdim(float x) { real.setxdim(x); imag.setxdim(x); }
    void setydim(float y) { real.setydim(y); imag.setydim(y); }
    void setzdim(float z) { real.setzdim(z); imag.setzdim(z); }
    void setdims(float x, float y, float z){ setxdim(x); setydim(y); setzdim(z); }
    int nvoxels() const { return real.nvoxels(); }
  
    const complexvolume& operator+=(const complexpoint& val);
    const complexvolume& operator-=(const complexpoint& val);
    const complexvolume& operator*=(const complexpoint& val);
    const complexvolume& operator/=(const complexpoint& val);
    const complexvolume& operator+=(const complexvolume& source); 
    const complexvolume& operator-=(const complexvolume& source); 
    const complexvolume& operator*=(const complexvolume& source); 
    const complexvolume& operator/=(const complexvolume& source); 
    complexvolume operator+(const complexpoint& val) const;
    complexvolume operator-(const complexpoint& val) const;
    complexvolume operator*(const complexpoint& val) const;
    complexvolume operator/(const complexpoint& val) const;
    complexvolume operator+(const complexvolume& vol) const;
    complexvolume operator-(const complexvolume& vol) const;
    complexvolume operator*(const complexvolume& vol) const;
    complexvolume operator/(const complexvolume& vol) const;

    inline const std::vector<int>& limits() const { return real.limits(); }
    inline int limits(int n) const { return real.limits(n); }
    inline int minx() const { return real.minx(); }
    inline int maxx() const { return real.maxx(); }
    inline int miny() const { return real.miny(); }
    inline int maxy() const { return real.maxy(); }
    inline int minz() const { return real.minz(); }
    inline int maxz() const { return real.maxz(); }
    inline const std::vector<int>& ROIlimits() const { return real.ROIlimits(); }
    inline int ROIlimits(int n) const { return real.ROIlimits(n); }
    inline bool usingROI() const { return real.usingROI(); }
    inline void setROIlimits(int x0, int y0, int z0, int x1, int y1, int z1) const
      { real.setROIlimits(x0,y0,z0,x1,y1,z1);
      imag.setROIlimits(x0,y0,z0,x1,y1,z1); }
    inline void setROIlimits(const std::vector<int>& lims) const
      { real.setROIlimits(lims);imag.setROIlimits(lims); }
    inline void activateROI() const
      { real.activateROI(); imag.activateROI(); } 
    inline void deactivateROI() const 
      { real.deactivateROI(); imag.deactivateROI(); }

    volume<float> abs() const;
    volume<float> phase() const;
    volume<float> re() const;
    volume<float> im() const;

    complexref operator()(int x,int y, int z)
      { return(complexref(&real(x,y,z),&imag(x,y,z))); }

    
    complexvolume extract_slice(int slice) const;
    void overwrite_slice(const complexvolume& data,int slice);

  };
}
#endif
