// General IO functions (images and transformation files)
//  Written by Mark Jenkinson  8/9/99

#include "newimageio.h"

using namespace MISCMATHS;

namespace NEWIMAGE {



////////////////////////////////////////////////////////////////////////////

// VOLUME I/O

int make_basename(string& filename)
{
  Tracer tr("make_basename");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  int fsize = basename.length(), indx;
  // find either .hdr or .img extensions and remove them
  indx=basename.rfind(".hdr");
  if ((indx>=0) && (indx<fsize)) {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  indx=basename.rfind(".img");
  if ((indx>=0) && (indx<fsize)) {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  indx = basename.length()-1;
  if (basename[indx] == '.') {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  // this is now the basename
  filename = basename;
  return 0;
}


int find_pathname(string& filename)
{
  Tracer tr("find_pathname");
  if (filename.size() < 1) return -1;
  string pathname = filename;
  int fsize = pathname.length(), indx;

  // working backwards, find '/' and remove everything after it

  indx = fsize;
  while ((pathname[indx] != '/') && (indx != 0))
    indx--;
  
  pathname.erase(indx+1,fsize);
  
  filename = pathname;
  return 0;
}


void check_filename(const string& basename)
{
  string fname;
  // check that the .hdr file exists
  fname = basename + ".hdr";
  ifstream ftest1(fname.c_str());
  if (!ftest1) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest1.close();

  // check that the .img file exists
  fname = basename + ".img";
  ifstream ftest2(fname.c_str());
  if (!ftest2) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest2.close();
}


AVW* NewAvwOpen(const string& filename, const string& permissions, 
		const volumeinfo& vinfo, bool use_vinfo)
{
  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) {
    return 0;
  }
  if (permissions=="r") check_filename(basename);

  AVW* OP=AvwOpen(basename.c_str(),permissions.c_str());
  if (OP==0) {
    cerr << "ERROR: Could not open image " << basename << endl;
    return 0;
  }
  if (use_vinfo) AvwCloneHeader(OP,&vinfo);  
  return OP;
}

AVW* NewAvwOpen(const string& filename, const string& permissions)
{
  volumeinfo vinfo;
  return NewAvwOpen(filename,permissions,vinfo,false);
}



short dtype(const char* T)   { return DT_UNSIGNED_CHAR; }
short dtype(const short* T)  { return DT_SIGNED_SHORT; }
short dtype(const int* T)    { return DT_SIGNED_INT; }
short dtype(const float* T)  { return DT_FLOAT; }
short dtype(const double* T) { return DT_DOUBLE; }

short dtype(const volume<char>& vol)   { return DT_UNSIGNED_CHAR; }
short dtype(const volume<short>& vol)  { return DT_SIGNED_SHORT; }
short dtype(const volume<int>& vol)    { return DT_SIGNED_INT; }
short dtype(const volume<float>& vol)  { return DT_FLOAT; }
short dtype(const volume<double>& vol) { return DT_DOUBLE; }

short dtype(const volume4D<char>& vol)   { return DT_UNSIGNED_CHAR; }
short dtype(const volume4D<short>& vol)  { return DT_SIGNED_SHORT; }
short dtype(const volume4D<int>& vol)    { return DT_SIGNED_INT; }
short dtype(const volume4D<float>& vol)  { return DT_FLOAT; }
short dtype(const volume4D<double>& vol) { return DT_DOUBLE; }

short dtype(const string& filename) 
{
  Tracer trcr("dtype");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");

  short dtype;
  AvwGetDataType(IP1,&dtype);
  AvwClose(IP1);

  return dtype;
}


volumeinfo volinfo(const string& filename) 
{
  Tracer trcr("volinfo");
  volumeinfo vinfo;
  if ( filename.size()<1 ) return vinfo;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");

  AvwCloneHeader(&vinfo,IP1);
  AvwClose(IP1);

  return vinfo;
}

//////////////////////////////////////////////////////////////////////////

// COMPLEX IMAGE I/O

void AvwReadComplexBuffer(AVW* IP, float* realbuffer, float* imagbuffer) 
{
  short sx,sy,sz,st;
  AvwGetDim(IP,&sx,&sy,&sz,&st);
  size_t imagesize=sx*sy*sz;
  short type;
  AvwGetDataType(IP,&type);
  switch(type)
    {
    case DT_COMPLEX:
      {
	float* sbuffer=new float[2*imagesize];
	if (sbuffer==0) { imthrow("Out of memory",99); }
	AvwReadVolumes(IP,sbuffer,1);
	float *sptr=sbuffer, *rptr=realbuffer, *iptr=imagbuffer;
	for (size_t poff=0; poff<imagesize; poff++) {
	  *rptr++ = *sptr++;
	  *iptr++ = *sptr++;
	}
	delete[] sbuffer;
      }
      break;
    default:
      ostrstream errmsg;
      errmsg << "AvwReadComplex: DT " << type <<  " is not complex!";
      perror(errmsg.str());
    }
}
  
void AvwWriteComplexVolume(AVW* OP, const float* realbuffer,
			    const float* imagbuffer) 
{
  short sx,sy,sz,st;
  AvwGetDim(OP,&sx,&sy,&sz,&st);
  size_t imagesize=sx*sy*sz*1;
  float* sbuffer=new float[2*imagesize];
  if (sbuffer==0) { imthrow("Out of memory",99); }
  float *sptr=sbuffer;
  const float *rptr=realbuffer, *iptr=imagbuffer;
  for (size_t poff=0; poff<imagesize; poff++) {
    *sptr++ = *rptr++;
    *sptr++ = *iptr++;
  }
  AvwWriteVolumes(OP,sbuffer,1);
  delete[] sbuffer;
}


int read_complexvolume(volume<float>& realvol, volume<float>& imagvol,
		       const string& filename, volumeinfo& vinfo,
		       bool read_img_data)
{
  Tracer trcr("read_complexvolume");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");
  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  size_t volsize=sx*sy*sz;

  float* realbuffer=new float[volsize];
  if (realbuffer==0) { imthrow("Out of memory",99); }
  float* imagbuffer=new float[volsize];
  if (imagbuffer==0) { imthrow("Out of memory",99); }
  if (read_img_data)  AvwReadComplexBuffer(IP1,realbuffer,imagbuffer);
  realvol.reinitialize(sx,sy,sz,realbuffer,true);
  imagvol.reinitialize(sx,sy,sz,imagbuffer,true);

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  realvol.setdims(x,y,z);
  imagvol.setdims(x,y,z);

  load_avw_origin(realvol,IP1);
  load_avw_origin(imagvol,IP1);

  AvwCloneHeader(&vinfo,IP1);
  AvwClose(IP1);
  return 0;
}


int read_complexvolume(volume<float>& realvol, volume<float>& imagvol,
		       const string& filename)
{
  volumeinfo vinfo;
  int retval = read_complexvolume(realvol,imagvol,filename,vinfo,true);
  return retval;
}


int read_complexvolume4D(volume4D<float>& realvols, volume4D<float>& imagvols,
			 const string& filename, volumeinfo& vinfo,
			 bool read_img_data)
{
  Tracer trcr("read_complexvolume4D");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");
  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  size_t volsize=sx*sy*sz;

  volume<float> dummyvol(sx,sy,sz);
  for (int t=0; t<st; t++) {
    realvols.addvolume(dummyvol);
    imagvols.addvolume(dummyvol);
    float* rbuffer=new float[volsize];
    if (rbuffer==0) { imthrow("Out of memory",99); }
    float* ibuffer=new float[volsize];
    if (ibuffer==0) { imthrow("Out of memory",99); }
    if (read_img_data)  AvwReadComplexBuffer(IP1,rbuffer,ibuffer);
    // Note that the d_owner flag = true in the following so that the
    //  control for delete is passed to the volume class
    realvols[t].reinitialize(sx,sy,sz,rbuffer,true);
    imagvols[t].reinitialize(sx,sy,sz,ibuffer,true);
  }

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  realvols.setdims(x,y,z,tr);
  imagvols.setdims(x,y,z,tr);

  for (int t=0; t<realvols.tsize(); t++) {
    load_avw_origin(realvols[t],IP1);
    load_avw_origin(imagvols[t],IP1);
  }

  AvwCloneHeader(&vinfo,IP1);
  AvwClose(IP1);
  return 0;
}


int read_complexvolume4D(volume4D<float>& realvol, volume4D<float>& imagvol,
			 const string& filename)
{
  volumeinfo vinfo;
  int retval = read_complexvolume4D(realvol,imagvol,filename,vinfo,true);
  return retval;
}

int read_complexvolume4D(volume4D<float>& realvol, volume4D<float>& imagvol,
			 const string& filename, volumeinfo& vinfo)
{
  int retval = read_complexvolume4D(realvol,imagvol,filename,vinfo,true);
  return retval;
}


int save_complexvolume(const volume<float>& realvol, 
		       const volume<float>& imagvol, const string& filename,
		       const volumeinfo& vinfo, bool use_vinfo)
{
  Tracer tr("save_complexvolume");
  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) return -1;

  AVW* OP=AvwOpen(basename.c_str(),"w");
  if (OP==0) return -1;
  if (use_vinfo) AvwCloneHeader(OP,&vinfo);
    
  int imin, imax;
  imin = (int) realvol.min();
  imax = (int) realvol.max();
  if (imax<=imin)  imax=imin+1;
  AvwSetMinMax(OP,imin,imax);
  save_avw_origin(realvol,OP);
    
  AvwSetDim(OP,realvol.xsize(),realvol.ysize(),realvol.zsize(),1);
  AvwSetDataType(OP, DT_COMPLEX);
  AvwSetVoxDim(OP,realvol.xdim() * realvol.xsign(),
	       realvol.ydim() * realvol.ysign(),
	       realvol.zdim() * realvol.zsign(),1.0);
  
  AvwWriteComplexVolume(OP,&(realvol(0,0,0)),&(imagvol(0,0,0)));

  AvwClose(OP); 
  return 0;
}


int save_complexvolume(const volume<float>& realvol, 
		       const volume<float>& imagvol, const string& filename)
{  
  volumeinfo vinfo;
  return save_complexvolume(realvol,imagvol,filename,vinfo,false);
}


int save_complexvolume(const volume<float>& realvol, 
		       const volume<float>& imagvol, const string& filename,
		       const volumeinfo& vinfo)
{  
  return save_complexvolume(realvol,imagvol,filename,vinfo,true);
}


int save_complexvolume4D(const volume4D<float>& realvols, 
			 const volume4D<float>& imagvols, 
			 const string& filename,
			 const volumeinfo& vinfo, bool use_vinfo)
{
  Tracer tr("save_complexvolume4D");

  if (realvols.tsize()<=0) return -1;

  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) return -1;

  AVW* OP=AvwOpen(basename.c_str(),"w");
  if (OP==0) return -1;
  if (use_vinfo) AvwCloneHeader(OP,&vinfo);
    
  int imin, imax;
  imin = (int) realvols.min();
  imax = (int) realvols.max();
  if (imax<=imin)  imax=imin+1;
  AvwSetMinMax(OP,imin,imax);
  save_avw_origin(realvols[0],OP);
    
  AvwSetDim(OP,realvols.xsize(),realvols.ysize(),realvols.zsize(),
	    realvols.tsize());
  AvwSetDataType(OP, DT_COMPLEX);
  AvwSetVoxDim(OP,realvols.xdim() * realvols.xsign(),
	       realvols.ydim() * realvols.ysign(),
	       realvols.zdim() * realvols.zsign(),
	       realvols.tdim());
  
  for (int t=0; t<realvols.tsize(); t++) {
    AvwWriteComplexVolume(OP,&(realvols[t](0,0,0)),&(imagvols[t](0,0,0)));
  }

  AvwClose(OP); 
  return 0;
}


int save_complexvolume4D(const volume4D<float>& realvol, 
			 const volume4D<float>& imagvol, const string& filename)
{  
  volumeinfo vinfo;
  return save_complexvolume4D(realvol,imagvol,filename,vinfo,false);
}


int save_complexvolume4D(const volume4D<float>& realvol, 
			 const volume4D<float>& imagvol, const string& filename,
			 const volumeinfo& vinfo)
{  
  return save_complexvolume4D(realvol,imagvol,filename,vinfo,true);
}



//////////////////////////////////////////////////////////////////////////

// MATRIX I/O



int read_ascii_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("read_ascii_matrix");
  int maxi = target.Nrows(), maxj = target.Ncols();
  if (maxi<=0) maxi=4;
  if (maxj<=0) maxj=4;
  target.ReSize(maxi,maxj);
  for (int i=1; i<=maxi; i++) {
    for (int j=1; j<=maxj; j++) {
      matfile >> target(i,j);
    }
  }
  return 0;
}


int read_ascii_matrix(Matrix &target, const string& filename)
{
  Tracer tr("read_ascii_matrix");
  if ( filename.size()<1 ) return -1;
  ifstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open matrix file " << filename << endl;
    return -1;
  }
  int rval = read_ascii_matrix(target,matfile);
  matfile.close();
  return rval;
}


int get_medx_small_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_medx_small_matrix");
  string str1;
  matfile >> str1;
  if (str1 != "[") {
    return -1;
  }
  int i=1, j=1;
  matfile >> str1;
  while (str1 != "]") {
    target(i,j) = atof(str1.c_str());
    if (j==4) { i++; j=1; }
    else { j++; }
    matfile >> str1;
  }
  return 0;
}  


int get_medx_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_medx_matrix");
  string str1, str2;
  matfile >> str1 >> str2;
  if ((str1 != "<<") || (str2 != "/matrix")) {
    return -1;
  }
  target.ReSize(4,4);
  Identity(target);  // default return value
  return get_medx_small_matrix(target,matfile);
}  


int get_minc_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_minc_matrix");
  string str1;
  matfile >> str1;
  if (str1 != "=") {
    cerr << "Could not parse MINC transform file" << endl;
    return -1;
  }
  target.ReSize(4,4);
  Identity(target);  // default return value
  for (int i=1; i<=3; i++) {
    for (int j=1; j<=4; j++) {
      matfile >> str1;
      target(i,j) = atof(str1.c_str());
    }
  }
  return 0;
}  


//------------------------------------------------------------------------//

//  int write_ascii_matrix(const Matrix& affmat, const string& filename)
//  {
//    Tracer tr("write_ascii_matrix");
//    if ( (filename.size()<1) ) return -1;
//    ofstream matfile(filename.c_str());
//    if (!matfile) { 
//      cerr << "Could not open file " << filename << " for writing" << endl;
//      return -1;
//    }

//    for (int i=1; i<=affmat.Nrows(); i++) {
//      for (int j=1; j<=affmat.Ncols(); j++) {
//        matfile << affmat(i,j) << "  ";
//      }
//      matfile << endl;
//    }
//    matfile.close();
//    return 0;
//  }



int put_medx_matrix(ofstream& matfile, const string& name, const Matrix& affmat)
{
  Tracer tr("put_medx_matrix");
  if (affmat.Nrows()<=0) { return -1; }
  matfile << "        /" << name << " [" << endl;
  for (int i=1; i<=affmat.Nrows(); i++) {
    for (int j=1; j<=affmat.Ncols(); j++) {
      matfile << "            " << affmat(i,j) << endl;
    }
  }
  matfile << "        ]" << endl;
  return 0;
}


int get_outputusermat(const string& filename, Matrix& oumat)
{
  Tracer trcr("get_outputusermat");
  if ( filename.size()<1 ) return -1;
  string basename = filename, fname;
  make_basename(basename);

  // check that the .hdr file exists
  fname = basename + ".hdr";
  ifstream ftest1(fname.c_str());
  if (!ftest1) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest1.close();

  AVW* IP1(AvwOpen(basename.c_str(), "r"));
  short orig[5];
  AvwGetOriginator(IP1,orig);
  ColumnVector origin(3);
  origin(1) = orig[0];
  origin(2) = orig[1];
  origin(3) = orig[2];

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);

  // Convert the SPM origin (as stored in the .hdr) to MEDx origin
  origin = 1.0 - origin;

  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  origin(2) = -sy - origin(2);   // y coordinate is treated weirdly

  oumat.ReSize(4,4);
  Identity(oumat);
  oumat(1,1) = x;
  oumat(2,2) = -y;
  oumat(3,3) = z;

  oumat(1,4) = origin(1)*x;
  oumat(2,4) = -origin(2)*y;
  oumat(3,4) = origin(3)*z;

  AvwClose(IP1);
  return 0;
}  



}









